import json
from os import environ
from typing import Dict, Any, Optional, Union, List, Callable

from helixtelemetry.telemetry.factory.telemetry_factory import TelemetryFactory
from helixtelemetry.telemetry.spans.telemetry_span_creator import TelemetrySpanCreator
from helixtelemetry.telemetry.spans.telemetry_span_wrapper import TelemetrySpanWrapper
from helixtelemetry.telemetry.structures.telemetry_parent import TelemetryParent

from spark_pipeline_framework.mixins.loop_id_mixin import LoopIdMixin
from spark_pipeline_framework.mixins.telemetry_parent_mixin import TelemetryParentMixin
from spark_pipeline_framework.utilities.async_helper.v1.async_helper import AsyncHelper

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml import Transformer
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)
from spark_pipeline_framework.utilities.parallel_pipeline_executor.v1.parallel_pipeline_executor import (
    ParallelPipelineExecutor,
)
from spark_pipeline_framework.utilities.spark_data_frame_helpers import (
    create_empty_dataframe,
)


class FrameworkParallelExecutor(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        *,
        enable: Union[bool, Callable[[DataFrame], bool]] = True,
        enable_if_view_not_empty: Optional[
            Union[str, Callable[[Optional[str]], str]]
        ] = None,
        stages: Union[List[Transformer], Callable[[], List[Transformer]]],
        max_parallel_tasks: int = 5,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        """
        Runs the provided stages in parallel


        :param stages: list of transformers or a function that returns a list of transformers
        :param max_parallel_tasks: number of concurrent tasks to create
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        self.max_parallel_tasks: int = max_parallel_tasks

        self.stages: Union[List[Transformer], Callable[[], List[Transformer]]] = stages

        self.enable: Union[bool, Callable[[DataFrame], bool]] = enable

        self.enable_if_view_not_empty: Optional[
            Union[str, Callable[[Optional[str]], str]]
        ] = enable_if_view_not_empty

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        if progress_logger is not None:
            progress_logger.write_to_log(
                f"---- Starting parallel run: "
                + f"max_parallel_tasks: {self.max_parallel_tasks}, "
                + f" ---------"
            )

        enable = self.enable(df) if callable(self.enable) else self.enable
        view_enable_if_view_not_empty = (
            self.enable_if_view_not_empty(self.loop_id)
            if callable(self.enable_if_view_not_empty)
            else self.enable_if_view_not_empty
        )
        enable_if_view_not_empty = (
            (
                df.sparkSession.catalog.tableExists(view_enable_if_view_not_empty)
                and not df.sparkSession.table(view_enable_if_view_not_empty).isEmpty()
            )
            if view_enable_if_view_not_empty
            else True
        )
        if (enable or enable is None) and enable_if_view_not_empty:
            stages: List[Transformer] = (
                self.stages if not callable(self.stages) else self.stages()
            )
            if stages and len(stages) > 0:
                if self.max_parallel_tasks > 1:
                    AsyncHelper.run(
                        self._process_async(df, stages, progress_logger=progress_logger)
                    )
                else:
                    self._process(df, stages, progress_logger=progress_logger)
        else:
            if progress_logger is not None:
                progress_logger.write_to_log(
                    self.getName() or "FrameworkTransformerGroup",
                    f"Skipping stages because enable "
                    + f"{self.enable or self.enable_if_view_not_empty} did not evaluate to True",
                )

        return df

    async def _process_async(
        self,
        df: DataFrame,
        stages: List[Transformer],
        progress_logger: Optional[ProgressLogger],
    ) -> None:
        """
        Saves the output asynchronously
        :param progress_logger:
        :param stages:
        :param df:
        """
        logger = get_logger(__name__)
        logger.info("Started process_async")

        pipeline_executor: ParallelPipelineExecutor = ParallelPipelineExecutor(
            progress_logger=progress_logger, max_tasks=self.max_parallel_tasks
        )

        telemetry_span_creator: TelemetrySpanCreator = TelemetryFactory(
            telemetry_parent=self.telemetry_parent or TelemetryParent.get_null_parent()
        ).create_telemetry_span_creator(log_level=environ.get("LOGLEVEL"))

        for stage in stages:
            stage_name: str = (
                stage.getName() if hasattr(stage, "getName") else "Unknown Transformer"
            )

            telemetry_span: TelemetrySpanWrapper
            async with telemetry_span_creator.create_telemetry_span_async(
                name=stage_name,
                attributes={},
                telemetry_parent=self.telemetry_parent,
            ) as telemetry_span:

                if isinstance(stage, LoopIdMixin):
                    stage.set_loop_id(self.loop_id)
                if isinstance(stage, TelemetryParentMixin):
                    stage.set_telemetry_parent(
                        telemetry_parent=telemetry_span.create_child_telemetry_parent()
                    )

                assert (
                    stage_name
                ), f"name must be set on every transformer in FrameworkParallelExecutor: f{json.dumps(stage, default=str)}"
                # convert each stage into a list of stages, so it can run in parallel
                pipeline_executor.append(name=stage_name, list_of_stages=[stage])

        # use a new df everytime to avoid keeping data in memory too long
        df.unpersist(blocking=True)
        df = create_empty_dataframe(df.sparkSession)

        async for name, _ in pipeline_executor.transform_async(df, df.sparkSession):
            async with telemetry_span_creator.create_telemetry_span_async(
                name=name,
                attributes={},
                telemetry_parent=telemetry_span.create_child_telemetry_parent(),
            ) as telemetry_span:
                if name:
                    logger.info(f"Finished running parallel stage {name}")

        logger.info("Finished process_async")

    def _process(
        self,
        df: DataFrame,
        stages: List[Transformer],
        progress_logger: Optional[ProgressLogger],
    ) -> None:
        """
        Saves the output asynchronously
        :param progress_logger:
        :param stages:
        :param df:
        """
        logger = get_logger(__name__)
        logger.info("Started process_sync")

        for stage in stages:
            stage_name: str = (
                stage.getName() if hasattr(stage, "getName") else "Unknown Transformer"
            )
            if isinstance(stage, LoopIdMixin):
                stage.set_loop_id(self.loop_id)
            if isinstance(stage, TelemetryParentMixin):
                stage.set_telemetry_parent(telemetry_parent=self.telemetry_parent)
            assert (
                stage_name
            ), f"name must be set on every transformer in FrameworkParallelExecutor: f{json.dumps(stage, default=str)}"
            df = stage.transform(df)

        logger.info("Finished process_sync")

    def as_dict(self) -> Dict[str, Any]:
        return {
            **(super().as_dict()),
            "stages": (
                [s.as_dict() for s in self.stages]  # type: ignore
                if not callable(self.stages)
                else str(self.stages)
            ),
        }



from functools import wraps
from typing import Any, Callable, TypeVar, cast

F = TypeVar("F", bound=Callable)  # type: ignore


def capture_parameters(func: F) -> F:
    """
    A decorator that forces keyword arguments in the wrapped method
    and saves actual input keyword arguments in `_input_kwargs`.

    Notes
    -----
    Should only be used to wrap a method where first arg is `self`
    """

    @wraps(func)
    def wrapper(self: Any, *args: Any, **kwargs: Any) -> Any:
        if len(args) > 0:
            raise TypeError("Method %s forces keyword arguments." % func.__name__)
        if hasattr(self, "_input_kwargs"):
            self._input_kwargs.update(
                kwargs
            )  # add to existing since this is being used in a subclass
        else:
            self._input_kwargs = kwargs
        return func(self, **kwargs)

    return cast(F, wrapper)



import pytest

from spark_pipeline_framework.utilities.slack.slack_client import SlackClient


@pytest.mark.skip("Actually sends a message to slack")
def test_slack_client() -> None:
    slack_token = "put in token here to test"
    channel: str = "#helix_pipeline_notifications"
    bot_user_name = "Helix Pipeline Monitor"
    client = SlackClient(
        slack_token=slack_token, channel=channel, bot_user_name=bot_user_name
    )
    slack_info = f'There are *{19}* double images detected for *{22}* products. Please check the <https://{"foo"}.s3-eu-west-1.amazonaws.com/{"bar"}|Double Images Monitor>.'

    # blocks = [{
    #     "type": "section",
    #     "text": {
    #         "type": "mrkdwn",
    #         "text": ":check: The script has run successfully on the dev."
    #     }
    # }]

    response = client.post_message_to_slack(slack_info, blocks=None)
    print(response)
    response = client.post_message_to_slack("reply", blocks=None)
    print(response)



import csv
from io import StringIO
from typing import List, Dict, Any, Optional, cast, AsyncGenerator, Type

import aiohttp
from aiohttp import ClientTimeout
from helix_fhir_client_sdk.utilities.list_chunker import ListChunker

from spark_pipeline_framework.utilities.helix_geolocation.v2.utilities.address_parser import (
    AddressParser,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.vendor_response import (
    VendorResponse,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.census_standardizing_vendor_api_response import (
    CensusStandardizingVendorApiResponse,
    CensusStandardizingVendorAddressMatch,
    CensusStandardizingVendorTigerLine,
    CensusStandardizingVendorCoordinates,
    CensusStandardizingVendorAddressComponents,
    CensusStandardizingVendorInput,
    CensusStandardizingVendorBenchmark,
    CensusStandardizingVendorAddress,
)


class CensusStandardizingVendor(
    StandardizingVendor[CensusStandardizingVendorApiResponse]
):
    @classmethod
    def get_api_response_class(cls) -> Type[CensusStandardizingVendorApiResponse]:
        return CensusStandardizingVendorApiResponse

    def __init__(
        self,
        use_bulk_api: bool = True,
        batch_request_max_size: Optional[int] = None,
        timeout: int = 5 * 60,
    ) -> None:
        super().__init__(version="1")
        self._use_bulk_api: bool = use_bulk_api
        # The Census service has a limit of 10,000 addresses per batch
        # https://geocoding.geo.census.gov/geocoder/Geocoding_Services_API.html
        self._batch_request_max_size: Optional[int] = batch_request_max_size or 9000
        self._timeout: int = timeout

    @classmethod
    def get_vendor_name(cls) -> str:
        return "census"

    def batch_request_max_size(self) -> int:
        return 9000

    async def standardize_async(
        self, raw_addresses: List[RawAddress], max_requests: int = 100
    ) -> List[VendorResponse[CensusStandardizingVendorApiResponse]]:
        """
        returns the vendor specific response from the vendor
        """
        vendor_specific_addresses: List[CensusStandardizingVendorApiResponse] = []

        assert all(
            [r.get_id() is not None for r in raw_addresses]
        ), f"{self.get_vendor_name()} requires all addresses to have an id. {[r.to_dict for r in raw_addresses]}"

        if self._use_bulk_api:
            for chunk in ListChunker().divide_into_chunks(
                raw_addresses, self.batch_request_max_size()
            ):
                vendor_specific_address_list: List[CensusStandardizingVendorApiResponse]
                async for vendor_specific_address_list in self._bulk_api_call_async(
                    raw_addresses=chunk
                ):
                    vendor_specific_addresses.extend(vendor_specific_address_list)

            # verify that we got the right count
            assert len(vendor_specific_addresses) == len(raw_addresses), (
                f"Number of standardized addresses {len(vendor_specific_addresses)}"
                f" does not match number of raw addresses {len(raw_addresses)}"
                "in the batch"
            )
        else:
            for address in raw_addresses:
                async for vendor_specific_address in self._single_api_call_async(
                    one_line_address=address.to_str(),
                    raw_address=address,
                ):
                    if vendor_specific_address:
                        vendor_specific_addresses.append(vendor_specific_address)
                    else:
                        vendor_specific_addresses.append(
                            CensusStandardizingVendorApiResponse.from_standardized_address(
                                StandardizedAddress.from_raw_address(
                                    address, vendor_name=self.get_vendor_name()
                                )
                            )
                        )

            assert len(vendor_specific_addresses) == len(raw_addresses), (
                f"Number of standardized addresses {len(vendor_specific_addresses)} "
                f"does not match number of raw addresses {len(raw_addresses)}"
            )

        # Now convert to vendor response
        vendor_responses: List[VendorResponse[CensusStandardizingVendorApiResponse]] = (
            self._to_vendor_response(
                vendor_response=vendor_specific_addresses,
                raw_addresses=raw_addresses,
                vendor_name=self.get_vendor_name(),
                response_version=self.get_version(),
            )
        )

        assert len(vendor_responses) == len(raw_addresses), (
            f"Number of vendor responses {len(vendor_responses)} does not match number of raw addresses"
            f" {len(raw_addresses)} in the batch"
        )
        return vendor_responses

    async def _bulk_api_call_async(
        self, *, raw_addresses: List[RawAddress]
    ) -> AsyncGenerator[List[CensusStandardizingVendorApiResponse], None]:
        """
        Make a bulk API call to the vendor

        :param raw_addresses: List of addresses
        :return: List of responses
        """

        # Create a CSV file with the addresses
        file_contents = ""  # '"Unique ID", "Street address", "City", "State", "ZIP"'
        valid_raw_addresses = [r for r in raw_addresses if r.is_valid_for_geolocation()]
        for address in valid_raw_addresses:
            file_contents += (
                f'"{address.get_internal_id()}", "{address.line1}", "{address.city}",'
                f' "{address.state}", "{address.zipcode}"\n'
            )

        # Define the URL and parameters
        url = "https://geocoding.geo.census.gov/geocoder/locations/addressbatch"
        form_data = aiohttp.FormData()
        form_data.add_field("addressFile", file_contents, filename="localfile.csv")
        form_data.add_field("benchmark", "4")

        async with aiohttp.ClientSession() as session:
            timeout = ClientTimeout(total=self._timeout)
            async with session.post(url, data=form_data, timeout=timeout) as response:
                response_text: str = await response.text()
                assert (
                    response.status == 200
                ), f"Error in the Census API call. Response code: {response.status}: {response_text}"
                # Check if the request was successful
                if response.status == 200:
                    # Use StringIO to treat the response text as a file-like object
                    f = StringIO(response_text)

                    # Define fieldnames based on the expected output
                    fieldnames = [
                        "ID",
                        "Input Address",
                        "Match Status",
                        "Match Type",
                        "Matched Address",
                        "Coordinates",
                        "TIGER Line ID",
                        "Side",
                    ]

                    # Use DictReader to parse the CSV response
                    reader: csv.DictReader[Any] = csv.DictReader(
                        f, fieldnames=fieldnames, skipinitialspace=True, quotechar='"'
                    )

                    # Convert reader to a list of dictionaries
                    # noinspection PyTypeChecker
                    results: List[Dict[str, Any]] = [
                        row for row in reader if any(row.values())
                    ]
                    # Now parse the results
                    responses = [
                        self._parse_csv_response(r, raw_addresses=raw_addresses)
                        for r in results
                    ]
                    # Find any records that are missing in responses that are present in raw_addresses
                    missing_raw_addresses = [
                        r
                        for r in raw_addresses
                        if r.get_internal_id() not in [r.address_id for r in responses]
                    ]
                    if missing_raw_addresses:
                        responses.extend(
                            [
                                CensusStandardizingVendorApiResponse.from_standardized_address(
                                    StandardizedAddress.from_raw_address(
                                        raw_address=r,
                                        vendor_name=self.get_vendor_name(),
                                    )
                                )
                                for r in missing_raw_addresses
                            ]
                        )

                    assert len(responses) == len(raw_addresses), (
                        f"Number of standardized addresses {len(responses)} does not match"
                        f" number of raw addresses {len(raw_addresses)}"
                        f" in the response: {response_text}"
                        f"\nAll raw addresses:\n"
                        + "\n".join([r.to_json() for r in raw_addresses])
                    )
                else:
                    responses = [
                        CensusStandardizingVendorApiResponse.from_standardized_address(
                            StandardizedAddress.from_raw_address(
                                raw_address=r, vendor_name=self.get_vendor_name()
                            )
                        )
                        for r in raw_addresses
                    ]
                # sort the list, so it is in the same order as the raw_addresses.
                # Census API does not return list in same order
                matching_responses: List[CensusStandardizingVendorApiResponse] = [
                    self.find_matching_response_for_raw_address(
                        raw_address=raw_address, responses=responses
                    )
                    for raw_address in raw_addresses
                ]
                assert len(matching_responses) == len(raw_addresses), (
                    f"Number of matching standardized addresses {len(matching_responses)} does not match number of"
                    f"raw addresses {len(raw_addresses)} in the batch: {response_text}"
                )
                yield matching_responses

    @staticmethod
    def find_matching_response_for_raw_address(
        *,
        raw_address: RawAddress,
        responses: List[CensusStandardizingVendorApiResponse],
    ) -> CensusStandardizingVendorApiResponse:
        matching_responses = [
            r for r in responses if r.address_id == raw_address.get_internal_id()
        ]
        assert len(matching_responses) > 0, (
            f"Could not find matching response for raw address id:\n{raw_address.get_internal_id()}"
            f"\n{raw_address.to_json()}"
            f"\nAll responses:\n" + "\n".join([r.address_id for r in responses])
        )
        return matching_responses[0]

    # noinspection PyMethodMayBeStatic
    def _parse_csv_response(
        self,
        r: Dict[str, Any],
        raw_addresses: List[RawAddress],
        benchmark: str = "Public_AR_Current",
    ) -> CensusStandardizingVendorApiResponse:

        # Split the address
        parsed_address: RawAddress | None = (
            AddressParser.split_address(
                address=cast(str, r.get("Matched Address")), address_id=r["ID"]
            )
            if r.get("Matched Address")
            else None
        )
        if not r.get("Matched Address") or not parsed_address:
            return CensusStandardizingVendorApiResponse(
                address_id=r["ID"],
                addressMatches=None,
                input=CensusStandardizingVendorInput(
                    address=CensusStandardizingVendorAddress(
                        address=r["Input Address"]
                    ),
                    benchmark=CensusStandardizingVendorBenchmark(
                        benchmarkName=benchmark,
                        benchmarkDescription=None,
                        isDefault=True,
                        id=None,
                    ),
                ),
            )
        line1 = parsed_address.line1
        line2 = parsed_address.line2
        city = parsed_address.city
        state = parsed_address.state
        zipcode = parsed_address.zipcode
        latitude = r["Coordinates"].split(",")[1] if r.get("Coordinates") else ""
        longitude = r["Coordinates"].split(",")[0] if r.get("Coordinates") else ""
        return CensusStandardizingVendorApiResponse(
            address_id=r["ID"],
            input=CensusStandardizingVendorInput(
                address=CensusStandardizingVendorAddress(address=r["Input Address"]),
                benchmark=CensusStandardizingVendorBenchmark(
                    benchmarkName=benchmark,
                    benchmarkDescription=None,
                    isDefault=True,
                    id=None,
                ),
            ),
            addressMatches=[
                CensusStandardizingVendorAddressMatch(
                    tigerLine=CensusStandardizingVendorTigerLine(
                        side=r["Side"], tigerLineId=r["TIGER Line ID"]
                    ),
                    coordinates=CensusStandardizingVendorCoordinates(
                        x=float(longitude) if longitude else None,
                        y=float(latitude) if latitude else None,
                    ),
                    addressComponents=CensusStandardizingVendorAddressComponents(
                        zip=zipcode,
                        streetName=line1,
                        preType=None,
                        city=city,
                        preDirection=None,
                        suffixDirection=None,
                        fromAddress=None,
                        state=state,
                        suffixType=None,
                        toAddress=None,
                        suffixQualifier=None,
                        preQualifier=None,
                    ),
                    matchedAddress=r["Matched Address"],
                )
            ],
        )

    # noinspection PyMethodMayBeStatic
    async def _single_api_call_async(
        self, *, one_line_address: str, raw_address: RawAddress
    ) -> AsyncGenerator[CensusStandardizingVendorApiResponse, None]:
        """
        Make an API call to the vendor for a single address

        :param one_line_address: One line address
        :return: Response as a dictionary
        """

        if not raw_address.is_valid_for_geolocation():
            yield CensusStandardizingVendorApiResponse.from_standardized_address(
                StandardizedAddress.from_raw_address(
                    raw_address, vendor_name=self.get_vendor_name()
                )
            )
            return

        params = {
            "address": one_line_address,
            "benchmark": "Public_AR_Current",
            "vintage": "Current_Current",
            "format": "json",
        }

        # Make the request
        # https://geocoding.geo.census.gov/geocoder/Geocoding_Services_API.html
        # ex: "https://geocoding.geo.census.gov/geocoder/locations/onelineaddress?address=9000%20Franklin%20Square%20Dr.%2C%20Baltimore%2C%20MD%2021237&benchmark=Public_AR_Current&format=json"
        url = f"https://geocoding.geo.census.gov/geocoder/locations/onelineaddress"
        async with aiohttp.ClientSession() as session:
            async with session.get(url, params=params) as response:
                response_text = await response.text()
                assert (
                    response.status == 200
                ), f"Error in the Census API call. Response code: {response.status}: {response_text}"

                # Check the response
                if response.status != 200:
                    yield CensusStandardizingVendorApiResponse.from_standardized_address(
                        StandardizedAddress.from_raw_address(
                            raw_address, vendor_name=self.get_vendor_name()
                        )
                    )
                    return

                # Parse the response
                #
                response_json: Dict[str, Any] = await response.json()
                if "result" not in response_json:
                    yield CensusStandardizingVendorApiResponse.from_standardized_address(
                        StandardizedAddress.from_raw_address(
                            raw_address, vendor_name=self.get_vendor_name()
                        )
                    )
                else:
                    geolocation_response: (
                        CensusStandardizingVendorApiResponse | None
                    ) = self._parse_geolocation_response(
                        response_json=response_json,
                        record_id=raw_address.get_internal_id(),
                    )
                    assert geolocation_response is None or isinstance(
                        geolocation_response, CensusStandardizingVendorApiResponse
                    )
                    if geolocation_response is not None:
                        yield geolocation_response
                    else:
                        yield CensusStandardizingVendorApiResponse.from_standardized_address(
                            StandardizedAddress.from_raw_address(
                                raw_address, vendor_name=self.get_vendor_name()
                            )
                        )

    @staticmethod
    def _parse_geolocation_response(
        *, response_json: Dict[str, Any], record_id: Optional[str]
    ) -> Optional[CensusStandardizingVendorApiResponse]:
        """
        Parse the JSON response from the US Census API and return a standardized address object

        :param response_json: JSON response
        :param record_id: Record ID
        :return: Standardized FHIR Address object
        """
        if not response_json:
            return None

        # Example response
        # {
        #      "result": {
        #           "input": {
        #                "address": {
        #                     "address": "4600 Silver Hill Rd, Washington, DC 20233"
        #                },
        #                "benchmark": {
        #                     "isDefault": true,
        #                     "benchmarkDescription": "Public Address Ranges - Current Benchmark",
        #                     "id": "4",
        #                     "benchmarkName": "Public_AR_Current"
        #                }
        #           },
        #           "addressMatches": [
        #                {
        #                     "tigerLine": {
        #                          "side": "L",
        #                          "tigerLineId": "76355984"
        #                     },
        #                     "coordinates": {
        #                          "x": -76.92748724230096,
        #                          "y": 38.84601622386617
        #                     },
        #                     "addressComponents": {
        #                          "zip": "20233",
        #                          "streetName": "SILVER HILL",
        #                          "preType": "",
        #                          "city": "WASHINGTON",
        #                          "preDirection": "",
        #                          "suffixDirection": "",
        #                          "fromAddress": "4600",
        #                          "state": "DC",
        #                          "suffixType": "RD",
        #                          "toAddress": "4700",
        #                          "suffixQualifier": "",
        #                          "preQualifier": ""
        #                     },
        #                     "matchedAddress": "4600 SILVER HILL RD, WASHINGTON, DC, 20233"
        #                }
        #           ]
        #      }
        # }
        # Get the result
        result: Dict[str, Any] | None = cast(
            Dict[str, Any] | None, response_json.get("result")
        )
        if not result or "addressMatches" not in result:
            return None

        result["address_id"] = record_id
        return CensusStandardizingVendorApiResponse.from_dict(result)

    def vendor_specific_to_std(
        self,
        *,
        vendor_specific_addresses: List[
            VendorResponse[CensusStandardizingVendorApiResponse]
        ],
        raw_addresses: List[RawAddress],
    ) -> List[StandardizedAddress]:
        """
        Each vendor class knows how to convert its response to StdAddress
        """
        assert all([r.get_id() for r in raw_addresses])
        assert all(
            [r.related_raw_address is not None for r in vendor_specific_addresses]
        )
        assert all([r.related_raw_address.get_id() for r in vendor_specific_addresses])
        std_addresses = [
            a.api_call_response.to_standardized_address(
                address_id=(a.related_raw_address.get_id())
            )
            for a in vendor_specific_addresses
        ]
        # if no standardized address found then use raw_address
        std_addresses = [
            (
                r
                if r is not None
                else self.get_matching_address(
                    raw_addresses=raw_addresses, address_id=r.address_id
                )
            )
            for r in std_addresses
        ]
        return std_addresses

    def _to_vendor_response(
        self,
        vendor_response: List[CensusStandardizingVendorApiResponse],
        raw_addresses: List[RawAddress],
        vendor_name: str,
        response_version: str,
    ) -> List[VendorResponse[CensusStandardizingVendorApiResponse]]:
        # create the map
        id_response_map: Dict[str, RawAddress] = {
            a.get_internal_id(): a for a in raw_addresses
        }
        ids_not_found = [
            r.address_id
            for r in vendor_response
            if r.address_id not in id_response_map.keys()
        ]
        assert (
            not ids_not_found
        ), f"Could not find raw addresses for ids: {ids_not_found}\n{vendor_response}"
        # find and assign
        return [
            VendorResponse(
                api_call_response=r,
                related_raw_address=id_response_map[r.address_id],
                vendor_name=vendor_name,
                response_version=response_version,
            )
            for r in vendor_response
        ]



from typing import Dict, Any, Optional

from pydantic import BaseModel

from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)


class GeocodioStandardizingVendorApiResponse(BaseModel, BaseVendorApiResponse):
    class Config:
        extra = "ignore"

    def to_dict(self) -> Dict[str, Any]:
        return self.model_dump()

    RecordID: str
    line1: Optional[str]
    line2: Optional[str]
    city: Optional[str]
    state: Optional[str]
    zipcode: Optional[str]
    geocodio_Country: Optional[str]
    Latitude: Optional[float]
    Longitude: Optional[float]

    @classmethod
    def from_raw_address(
        cls, raw_address: RawAddress
    ) -> "GeocodioStandardizingVendorApiResponse":
        return GeocodioStandardizingVendorApiResponse(
            RecordID=raw_address.address_id,
            line1=raw_address.line1,
            line2=raw_address.line2,
            city=raw_address.city,
            state=raw_address.state,
            zipcode=raw_address.zipcode,
            geocodio_Country=raw_address.country,
            Latitude=None,
            Longitude=None,
        )

    def to_standardized_address(self, *, address_id: str) -> StandardizedAddress:
        return StandardizedAddress(
            address_id=address_id,
            line1=self.line1,
            line2=self.line2,
            city=self.city,
            county=None,
            state=self.state,
            zipcode=self.zipcode,
            country=self.geocodio_Country,
            latitude=str(self.Latitude) if self.Latitude else None,
            longitude=str(self.Longitude) if self.Longitude else None,
            formatted_address=f"{self.line1}, {self.city}, {self.state} {self.zipcode}",
            standardize_vendor="geocodio",
        )



from typing import Any, Dict, Optional

from spark_pipeline_framework.utilities.athena.athena import Athena
from spark_pipeline_framework.utilities.athena.athena_source_file_type import (
    AthenaSourceFileType,
)

# noinspection PyProtectedMember
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_log_metric import (
    ProgressLogMetric,
)
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class AthenaTableCreator(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        view: str,
        schema_name: str,
        table_name: str,
        s3_source_path: str,
        source_file_type: AthenaSourceFileType,
        s3_temp_folder: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ) -> None:
        """
        Creates an Athena table

        :param s3_temp_folder: where Athena should store temp files
        """
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        assert view

        self.logger = get_logger(__name__)

        self.view: Param[str] = Param(self, "view", "")
        self._setDefault(view=view)

        self.schema_name: Param[str] = Param(self, "schema_name", "")
        self._setDefault(schema_name=schema_name)

        self.table_name: Param[str] = Param(self, "table_name", "")
        self._setDefault(table_name=table_name)

        self.s3_source_path: Param[str] = Param(self, "s3_source_path", "")
        self._setDefault(s3_source_path=s3_source_path)

        self.source_file_type: Param[AthenaSourceFileType] = Param(
            self, "source_file_type", ""
        )
        self._setDefault(source_file_type=source_file_type)

        self.s3_temp_folder: Param[str] = Param(self, "s3_temp_folder", "")
        self._setDefault(s3_temp_folder=s3_temp_folder)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view: str = self.getView()
        schema_name: str = self.getSchemaName()
        table_name: str = self.getTableName()
        s3_source_path: str = self.getS3SourcePath()
        source_file_type: AthenaSourceFileType = self.getSourceFileType()
        name: Optional[str] = self.getName()
        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()
        s3_temp_folder: str = self.getS3TempFolder()

        assert s3_temp_folder, "s3_temp_folder must be set or we cannot write to Athena"

        self.logger.info(
            f"----- Started creating athena table for {name or view} --------"
        )
        with ProgressLogMetric(
            name=f"{name or view}_athena_table_creator", progress_logger=progress_logger
        ):
            json_df: DataFrame = df.sparkSession.table(view)
            try:
                Athena.drop_create_table(
                    table_name,
                    schema_name,
                    json_df,
                    s3_source_path,
                    source_file_type,
                    s3_temp_folder=s3_temp_folder,
                )
            except Exception as e:
                self.logger.warning(
                    f"Error writing athena table for {name or view}: {str(e)}"
                )
                raise

        self.logger.info(
            f"----- Finished creating athena table for {name or view} --------"
        )

        return df

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getView(self) -> str:
        return self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSchemaName(self) -> str:
        return self.getOrDefault(self.schema_name)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getTableName(self) -> str:
        return self.getOrDefault(self.table_name)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getS3SourcePath(self) -> str:
        return self.getOrDefault(self.s3_source_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getSourceFileType(self) -> AthenaSourceFileType:
        return self.getOrDefault(self.source_file_type)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getName(self) -> Optional[str]:
        return self.getOrDefault(self.name) or self.getOrDefault(self.view)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getS3TempFolder(self) -> str:
        return self.getOrDefault(self.s3_temp_folder)



import dataclasses
from typing import Optional

from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_udf_parameters import (
    AsyncPandasUdfParameters,
)


@dataclasses.dataclass
class ElasticSearchSenderParameters:
    total_partitions: int
    name: Optional[str]
    index: str
    operation: str
    doc_id_prefix: Optional[str]
    log_level: Optional[str]
    timeout: Optional[int]
    pandas_udf_parameters: AsyncPandasUdfParameters



from dataclasses import dataclass
from typing import Any, Dict, List


@dataclass
class ElasticSearchResult:
    url: str
    success: int
    failed: int
    payload: List[Dict[str, Any]]
    partition_index: int



from os import path, makedirs
from pathlib import Path
from shutil import rmtree

from pyspark.sql import SparkSession, DataFrame
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.column_name_cleaner.column_name_cleaner import (
    FrameworkColumnNameCleaner,
)


def test_column_name_cleaner(spark_session: SparkSession) -> None:
    # Arrange
    data_dir: Path = Path(__file__).parent.joinpath("./")

    temp_folder = data_dir.joinpath("./temp")
    if path.isdir(temp_folder):
        rmtree(temp_folder)
    makedirs(temp_folder)

    df: DataFrame = spark_session.read.json(str(data_dir.joinpath("input")))
    view = "nested_field_table"
    df.createOrReplaceTempView(view)

    df.printSchema()

    # Act
    with ProgressLogger() as progress_logger:
        FrameworkColumnNameCleaner(
            view=view, progress_logger=progress_logger
        ).transform(df)

    df = spark_session.table(view)

    df.printSchema()

    # Assert
    df.write.parquet(str(data_dir.joinpath("temp").joinpath("output")))



from datetime import datetime

import pytest
from aioresponses import aioresponses
from typing import Dict, List, Any, Optional
from unittest.mock import AsyncMock, MagicMock

from spark_pipeline_framework.transformers.address_standardization.v2.address_standardization_processor import (
    AddressStandardizationProcessor,
    AddressStandardizationParameters,
)
from spark_pipeline_framework.utilities.async_pandas_udf.v1.async_pandas_batch_function_run_context import (
    AsyncPandasBatchFunctionRunContext,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_handler import (
    CacheHandler,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.cache.cache_result import (
    CacheResult,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.raw_address import (
    RawAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.structures.standardized_address import (
    StandardizedAddress,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.standardizing_vendor import (
    StandardizingVendor,
)
from spark_pipeline_framework.utilities.helix_geolocation.v2.vendors.vendor_responses.base_vendor_api_response import (
    BaseVendorApiResponse,
)


@pytest.mark.asyncio
async def test_standardize_list() -> None:
    input_values: List[Dict[str, Any]] = [
        {
            "id": "1",
            "address1": "123 Main St",
            "address2": "",
            "city": "Anytown",
            "state": "CA",
            "zip": "12345",
        }
    ]

    # Mock StandardizingVendor and BaseVendorApiResponse
    mock_vendor_response = MagicMock(spec=BaseVendorApiResponse)
    mock_standardizing_vendor = MagicMock(spec=StandardizingVendor)
    mock_standardizing_vendor.standardize_async = AsyncMock(
        return_value=[mock_vendor_response]
    )
    mock_standardizing_vendor.vendor_specific_to_std = MagicMock(
        return_value=[
            StandardizedAddress(
                address_id="1",
                line1="123 Main St",
                line2="",
                city="Anytown",
                state="CA",
                zipcode="12345",
                latitude="37.7749",
                longitude="-122.4194",
                standardize_vendor="mock_standardizing_vendor",
            )
        ]
    )

    mock_cache_handler = AsyncMock(spec=CacheHandler)
    mock_cache_handler.check_cache = AsyncMock(
        return_value=CacheResult(
            found=[],
            not_found=[
                RawAddress(
                    address_id="1",
                    line1="123 Main St",
                    line2="",
                    city="Anytown",
                    state="CA",
                    zipcode="12345",
                )
            ],
        )
    )
    mock_cache_handler.save_to_cache = AsyncMock()

    address_standardization_parameters = AddressStandardizationParameters(
        total_partitions=1,
        address_column_mapping={
            "address_id": "id",
            "line1": "address1",
            "line2": "address2",
            "city": "city",
            "state": "state",
            "zipcode": "zip",
        },
        standardizing_vendor=mock_standardizing_vendor,
        cache_handler=mock_cache_handler,
        geolocation_column_prefix="geo_",
        log_level="INFO",
    )

    expected_output: List[Dict[str, Optional[str]]] = [
        {
            "geo_latitude": "37.7749",
            "geo_longitude": "-122.4194",
            "address1": "123 Main St",
            "address2": "",
            "city": "Anytown",
            "state": "CA",
            "zip": "12345",
        }
    ]

    with aioresponses() as m:
        # Mock the HTTP request to the standardization service
        m.post("http://mock_standardization_service", payload=expected_output)

        async for result in AddressStandardizationProcessor.standardize_list(
            run_context=AsyncPandasBatchFunctionRunContext(
                partition_index=0,
                chunk_index=0,
                chunk_input_range=range(1),
                partition_start_time=datetime.now(),
            ),
            input_values=input_values,
            parameters=address_standardization_parameters,
            additional_parameters=None,
        ):
            assert result == expected_output[0]


@pytest.mark.asyncio
async def test_standardize_list_missing_parameters() -> None:
    input_values: List[Dict[str, Any]] = [
        {
            "id": "1",
            "address1": "123 Main St",
            "address2": "",
            "city": "Anytown",
            "state": "CA",
            "zip": "12345",
        }
    ]

    with pytest.raises(AssertionError, match="parameters is required"):
        async for _ in AddressStandardizationProcessor.standardize_list(
            run_context=AsyncPandasBatchFunctionRunContext(
                partition_index=0,
                chunk_index=0,
                chunk_input_range=range(1),
                partition_start_time=datetime.now(),
            ),
            input_values=input_values,
            parameters=None,
            additional_parameters=None,
        ):
            pass


@pytest.mark.asyncio
async def test_standardize_list_invalid_input() -> None:
    # Mock StandardizingVendor and BaseVendorApiResponse
    mock_vendor_response = MagicMock(spec=BaseVendorApiResponse)
    mock_standardizing_vendor = MagicMock(spec=StandardizingVendor)
    mock_standardizing_vendor.standardize_async = AsyncMock(
        return_value=[mock_vendor_response]
    )
    mock_cache_handler = MagicMock(spec=CacheHandler)
    mock_cache_handler.check_cache = MagicMock(
        return_value=CacheResult(found=[], not_found=[])
    )
    mock_cache_handler.save_to_cache = MagicMock()

    address_standardization_parameters = AddressStandardizationParameters(
        total_partitions=1,
        address_column_mapping=None,  # type: ignore[arg-type]
        standardizing_vendor=mock_standardizing_vendor,
        cache_handler=mock_cache_handler,
        geolocation_column_prefix="geo_",
        log_level="INFO",
    )

    input_values: List[Dict[str, Any]] = [
        {
            "id": None,
            "address1": "123 Main St",
            "address2": "",
            "city": "Anytown",
            "state": "CA",
            "zip": "12345",
        }
    ]

    with pytest.raises(AssertionError, match="address_column_mapping is required"):
        async for _ in AddressStandardizationProcessor.standardize_list(
            run_context=AsyncPandasBatchFunctionRunContext(
                partition_index=0,
                chunk_index=0,
                chunk_input_range=range(1),
                partition_start_time=datetime.now(),
            ),
            input_values=input_values,
            parameters=address_standardization_parameters,
            additional_parameters=None,
        ):
            pass



import json
from typing import List


def remove_field_from_json(json_str: str, field_list: List[str]) -> str:
    """
    Map function which removes given field from JSON
    """
    json_dict = json.loads(json_str)
    for field in field_list:
        del json_dict[field]
    return str(json.dumps(json_dict))



from importlib import import_module
from typing import Any, Dict, Optional, cast

from spark_pipeline_framework.utilities.capture_parameters import capture_parameters
from pyspark.ml.param import Param
from pyspark.sql.dataframe import DataFrame
from pyspark.sql.functions import from_json
from pyspark.sql.types import StructType

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger
from spark_pipeline_framework.transformers.framework_transformer.v1.framework_transformer import (
    FrameworkTransformer,
)


class ResourceConverter(FrameworkTransformer):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        # add your parameters here (be sure to add them to setParams below too)
        view_source: str,
        view_target: str,
        resource_type_name: str,
        name: Optional[str] = None,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
    ):
        super().__init__(
            name=name, parameters=parameters, progress_logger=progress_logger
        )

        self.logger = get_logger(__name__)

        assert view_source
        assert view_target
        assert resource_type_name

        # add params
        self.view_source: Param[str] = Param(self, "view_source", "")
        self._setDefault(view_source=view_source)

        self.view_target: Param[str] = Param(self, "view_target", "")
        self._setDefault(view_target=view_target)

        self.resource_type_name: Param[str] = Param(self, "resource_type_name", "")
        self._setDefault(resource_type_name=resource_type_name)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    def _transform(self, df: DataFrame) -> DataFrame:
        view_source: str = self.getViewSource()
        view_target: str = self.getViewTarget()
        resource_type_name: str = self.getResourceTypeName()

        progress_logger: Optional[ProgressLogger] = self.getProgressLogger()

        if progress_logger:
            progress_logger.write_to_log(
                name="ResourceConverter",
                message="Converting string json to StructType resource...",
            )

        df_source: DataFrame = df.sparkSession.table(view_source)

        df_source = df_source.withColumn(
            colName="struct_type_resource",
            col=from_json(
                col=df_source.resource,
                schema=self.get_schema_for_resource(
                    resource_type_name=resource_type_name
                ),
            ),
        )

        df_target: DataFrame = df_source.select("struct_type_resource")

        resource_id: Any = df_target.select("struct_type_resource.id").collect()[0][0]

        if progress_logger:
            progress_logger.write_to_log(
                name="ResourceConverter",
                message=f"Resources converted: id={resource_id}",
            )

        df_target.createOrReplaceTempView(name=view_target)

        return df_target

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getViewSource(self) -> str:
        return self.getOrDefault(self.view_source)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getViewTarget(self) -> str:
        return self.getOrDefault(self.view_target)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getResourceTypeName(self) -> str:
        return self.getOrDefault(self.resource_type_name)

    # Dynamic Module loading for the Schema classes for resources,
    #   based on the given resource types available
    def _get_r4_schema_class_path(self, fhir_resource_type_name: str) -> str:
        return f"spark_fhir_schemas.r4.resources.{fhir_resource_type_name.lower()}"

    def _get_r4_schema_class(self, resource_type_name: str) -> Any:
        try:
            module: Any = import_module(
                name=self._get_r4_schema_class_path(resource_type_name)
            )
            return getattr(module, f"{resource_type_name}Schema")
        except ModuleNotFoundError:
            raise NotImplementedError(
                f"A R4 Schema Class for {resource_type_name} resource type is not currently supported."
            )

    def get_schema_for_resource(self, resource_type_name: str) -> StructType:
        schema_class: Any = self._get_r4_schema_class(resource_type_name)
        schema: StructType = cast(StructType, schema_class.get_schema())
        return schema



import math
from typing import Optional

from pyspark.sql import DataFrame
from pyspark.sql.functions import get_json_object, col


class SparkPartitionHelper:
    @staticmethod
    def partition_if_needed(
        *,
        df: DataFrame,
        desired_partitions: int,
        enable_repartitioning: Optional[bool] = None,
        force_partition: Optional[bool] = None,
        partition_by_column_name: Optional[str] = None,
        enable_coalesce: Optional[bool] = None,
    ) -> DataFrame:
        """
        This function is used to partition the incoming dataframe based on the number of partitions or the partition size

        :param df: incoming dataframe
        :param desired_partitions: desired number of partitions
        :param enable_repartitioning: If this is not true then the incoming dataframe will not be repartitioned
        :param partition_by_column_name: column_name to partition by
        :param force_partition: If this is true then the incoming dataframe will be repartitioned even
                                if the number of partitions is the same
        :param enable_coalesce: If this is true then the incoming dataframe will be coalesced if
                                it has more partitions than desired
                                https://medium.com/@ashwin_kumar_/spark-repartition-vs-coalesce-034d748aab2e
        :return: repartitioned dataframe (if it was needed)
        """
        # see if we need to partition the incoming dataframe
        if enable_repartitioning:
            if (
                partition_by_column_name
            ):  # repartition even if the number of partitions is the same
                df = df.withColumn(
                    partition_by_column_name,
                    get_json_object(col("value"), f"$.{partition_by_column_name}"),
                ).repartition(desired_partitions, partition_by_column_name)
                df = df.drop(partition_by_column_name)
            elif (
                not force_partition and desired_partitions != df.rdd.getNumPartitions()
            ):
                if enable_coalesce and desired_partitions < df.rdd.getNumPartitions():
                    df = df.coalesce(desired_partitions)
                else:
                    df = df.repartition(desired_partitions)

        return df

    @staticmethod
    def calculate_desired_partitions(
        *, df: DataFrame, num_partitions: Optional[int], partition_size: Optional[int]
    ) -> int:
        """
        Calculates the desired number of partitions based on the incoming dataframe rows, number of partitions requested
        or partition size requested

        :param df: incoming dataframe
        :param num_partitions: desired number of partitions
        :param partition_size: desired partition size
        :return: number of partitions that we should repartition the incoming dataframe to that meets the request
        """
        row_count: int = df.count()
        if num_partitions:
            return num_partitions
        elif partition_size:
            return math.ceil(row_count / partition_size)
        else:
            return df.rdd.getNumPartitions()



from pathlib import Path
from typing import Any, Dict, Optional, Union, Callable

from pyspark.ml.param import Param

from spark_pipeline_framework.logger.yarn_logger import get_logger
from spark_pipeline_framework.progress_logger.progress_logger import ProgressLogger

from spark_pipeline_framework.transformers.framework_base_exporter.v1.framework_base_exporter import (
    FrameworkBaseExporter,
)
from spark_pipeline_framework.utilities.file_modes import FileWriteModes
from spark_pipeline_framework.utilities.capture_parameters import capture_parameters


class FrameworkCsvExporter(FrameworkBaseExporter):
    # noinspection PyUnusedLocal
    @capture_parameters
    def __init__(
        self,
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]],
        header: bool,
        delimiter: str = ",",
        view: Optional[str] = None,
        name: Optional[str] = None,
        mode: str = FileWriteModes.MODE_ERROR,
        parameters: Optional[Dict[str, Any]] = None,
        progress_logger: Optional[ProgressLogger] = None,
        limit: Optional[int] = None,
        stream: bool = False,
        delta_lake_table: Optional[str] = None,
    ):
        super().__init__(
            view=view,
            name=name,
            mode=mode,
            parameters=parameters,
            progress_logger=progress_logger,
            limit=limit,
            stream=stream,
            delta_lake_table=delta_lake_table,
        )

        assert (
            isinstance(file_path, Path)
            or isinstance(file_path, str)
            or callable(file_path)
        ), type(file_path)
        assert file_path

        self.logger = get_logger(__name__)

        self.file_path: Param[
            Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]
        ] = Param(self, "file_path", "")
        self._setDefault(file_path=None)

        self.header: Param[bool] = Param(self, "header", "")
        self._setDefault(header=header)

        self.delimiter: Param[str] = Param(self, "delimiter", "")
        self._setDefault(delimiter=delimiter)

        kwargs = self._input_kwargs
        self.setParams(**kwargs)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getFilePath(
        self,
    ) -> Union[Path, str, Callable[[Optional[str]], Union[Path, str]]]:
        return self.getOrDefault(self.file_path)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getHeader(self) -> bool:
        return self.getOrDefault(self.header)

    # noinspection PyPep8Naming,PyMissingOrEmptyDocstring
    def getDelimiter(self) -> str:
        return self.getOrDefault(self.delimiter)

    def getFormat(self) -> str:
        return "csv"

    def getOptions(self) -> Dict[str, Any]:
        file_path: Union[Path, str, Callable[[Optional[str]], Union[Path, str]]] = (
            self.getFilePath()
        )
        if callable(file_path):
            file_path = file_path(self.loop_id)
        return {
            "path": file_path,
            "header": self.getHeader(),
            "delimiter": self.getDelimiter(),
        }
